#!/usr/bin/env python

from urlparse import urlparse, parse_qs
import json
import re
import sys
import time
import threading
import urllib2
import uwsgi
import yaml

def encodeUserData(user, password):
    return "Basic " + (user + ":" + password).encode("base64").rstrip()

def getValueFromString(string, value):
    return re.search('(?<=%s=")[^"]*' % value, string).group(0)

class SearchThread(threading.Thread):
    def __init__(self, registry_url, user, password):
        threading.Thread.__init__(self)
        self.registry_url = registry_url
        self.user = user
        self.password = password
        self.entries = {}

    def dump(self):
        return self.entries

    def run(self):
        while True:
            try:
                self.updateCache()
            except:
                print 'Failed to update cache!'
            time.sleep(30)

    def updateCache(self):
        entries = {}
        # get registry entries
        catalog_suffix = '/v2/_catalog'
        token_request = self.getData(catalog_suffix)
        token = self.getToken(token_request, self.user, self.password)
        response = json.loads(self.getData(catalog_suffix, token))
        repositories = response['repositories']
        # get tags
        for repository in repositories:
            repository_suffix = '/v2/%s/tags/list' % repository
            token_request = self.getData(repository_suffix)
            token = self.getToken(token_request, self.user, self.password)
            data = self.getData(repository_suffix, token)
            # if no data about repository returned
            if not data:
                continue
            response = json.loads(data)
            entries[repository]=response['tags']
        # update entries registry
        self.entries = entries

    def getData(self, suffix, token=None):
        req = urllib2.Request('%s%s' % (self.registry_url, suffix))
        if token:
            req.add_header('Authorization', 'Bearer %s' % token)
        try:
            res = urllib2.urlopen(req)
            return res.read()
        except IOError, error_code:
            for header in str(error_code.headers).split('\n'):
                if header.startswith('Www-Authenticate'):
                    return {'realm': getValueFromString(header, 'realm'),
                            'service': getValueFromString(header, 'service'),
                            'scope': getValueFromString(header, 'scope'),}

    def getToken(self, token_request, user, password):
        url = '%s?service=%s&scope=%s' % (token_request['realm'],
                                           token_request['service'],
                                           token_request['scope'])
        req = urllib2.Request(url)
        req.add_header('Authorization', encodeUserData(user, password))
        res = urllib2.urlopen(req)
        return re.search('(?<=token":")[^"]*', res.read()).group(0)

    def lookupCache(self, name_match, tag_match):
        results = {}
        # try to get data, return empty if no entries yet
        for repository in self.entries.keys():
            repository_tmp = []
            if re.match(name_match, repository):
                if self.entries[repository]:
                    for tag in self.entries[repository]:
                        if re.match(tag_match, tag):
                            repository_tmp.append(tag)
                if repository_tmp:
                    results[repository] = repository_tmp
        return results


def initialize():
    global searchapi

    # read config file
    config_file = open('/etc/docker/registry/search.yml', 'r')
    config = yaml.load(config_file)['server']

    # initialize search thread
    searchapi = SearchThread(registry_url=config['registry_url'],
                             user=config['user'],
                             password=config['password'])
    searchapi.daemon = True
    searchapi.start()


def application(env, start_response):
    global searchapi

    parameters = parse_qs(env['QUERY_STRING'])

    # try to get parameters
    try:
        name = parameters['name'][0]
        tag = parameters['tag'][0]
        format = parameters['format'][0]
    except:
        start_response('500 INTERNAL SERVER ERROR',
                       [('Content-Type','text/plain')])
        return

    # prepare response
    search_results = searchapi.lookupCache(name, tag)
    if format == 'dictionary':
        message = str(search_results)
    elif format == 'select':
        message = ''
        for key in search_results.keys():
            for entry in search_results[key]:
                message += '%s:%s\n' % (key, entry)
    else:
        start_response('404 NOT FOUND', [('Content-Type','text/plain')])
        return

    # send response
    start_response('200 OK', [('Content-Type','text/plain')])
    return str(message)

uwsgi.post_fork_hook = initialize
